#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#define PS_PORT (PORTB)
#define PS_DDR (DDRB)
#define PS_PIN (PINB)

/* DATA: controller->console (MOSI) */
#define PS_DATA (3)
/* CMD: console->controller (MISO) */
#define PS_CMD (4)
/* ATT: console->controller (/SS) */
#define PS_ATT (2)
/* CLK: console->controller (SCK) */
#define PS_CLK (5)
/* ACK: controller->console */
#define PS_ACK (1)

volatile enum transfer_state {
	TRANSFER_STATE_IDLE,
	TRANSFER_STATE_AWAITING_COMMAND,
	TRANSFER_STATE_SENDING_DATA,
	TRANSFER_STATE_FINISHING,
} state;

volatile uint8_t command;
volatile uint8_t data_to_send;
volatile uint8_t counter = 0;

// scratch variables
uint8_t ix;
uint8_t mask;

// D.7 = [] / X
// D.6 = >< / A
// D.5 = () / B
// D.4 = /\ / Y
// D.3 = R1 / Rb
// D.2 = L1 / Lb
// D.1 = R2 / Rt
// D.0 = L2 / Lt
uint8_t button_mapping[8];

uint8_t remap_counter = 0;

void spi_enable(void) {
	uint8_t dummy;
	
	PS_PORT |= _BV(PS_ATT); // Weak pull-up on ATT (/SS)
	SPCR = _BV(SPIE) | _BV(SPE) | _BV(DORD) | _BV(CPOL) | _BV(CPHA);
	
	PS_DDR |= _BV(PS_CMD); // Make CMD (MISO) an output
	
	SPDR = 0xFF;
	dummy = SPSR;
	dummy = SPDR;
}

void spi_disable(void) {
	uint8_t dummy;
	
	PS_DDR &= ~_BV(PS_CMD);
	SPCR &= ~_BV(SPIE);
	dummy = SPSR;
	dummy = SPDR;
}

ISR(SPI_STC_vect) {
	
	uint8_t data = SPDR;
	uint8_t ack = !0;
	
	SPDR = 0xFF;
	
	switch (state) {
		case TRANSFER_STATE_IDLE:
			switch (data) {
				case 0x01:
					state = TRANSFER_STATE_AWAITING_COMMAND;
					SPDR = 0x41; // Digital pad.
					break;
				default:
					spi_disable();
					state = TRANSFER_STATE_IDLE;
					ack = 0;
					break;
			}
			break;
		case TRANSFER_STATE_AWAITING_COMMAND:
			command = data;
			++counter;
			switch (command) {
				case 0x42:
					state = TRANSFER_STATE_SENDING_DATA;
					data_to_send = 2; // 2 bytes of data.
					SPDR = 0x5A; // Data coming.
					break;
				default:
					state = TRANSFER_STATE_IDLE;
					ack = 0;
					break;
			}
			break;
		case TRANSFER_STATE_SENDING_DATA:
			switch (2 - data_to_send) {
				case 0:
					data = PINC;
					// WE HAVE: --XXXXXX
					// WE WANT: XXXXX--X
					SPDR = (data << 2) | 0b00000110 | (data & 0b00000001);
					break;
				case 1:
					data = 0xFF;
					for (ix = 0, mask = 1; mask; ++ix, mask <<= 1) {
						if (!(PIND & mask)) {
							data ^= button_mapping[ix];
						}
					}
					SPDR = data;
					break;
			}
			if (--data_to_send == 0) {
				state = TRANSFER_STATE_FINISHING;
			}
			break;
		case TRANSFER_STATE_FINISHING:
			state = TRANSFER_STATE_IDLE;
			ack = 0;
			break;
	}
	
	if (ack && !(PS_PIN & _BV(PS_ATT))) {
		PS_DDR |= _BV(PS_ACK);
		_delay_us(1);
		PS_DDR &= ~_BV(PS_ACK);
	} else {
		state = TRANSFER_STATE_IDLE;
		if (!(SPCR & _BV(SPIE))) {
			while (!(PS_PIN & _BV(PS_ATT)));
			spi_enable();
		}
	}
}

int main(void) {
	
	spi_enable();
	
	PORTC |= 0xFF;
	PORTD |= 0xFF;

	for (ix = 0, mask = 1; mask; ++ix, mask <<= 1) {
		button_mapping[ix] = mask;
	}
	
	sei();
	
	for (;;) {
		
		_delay_ms(10);
		
		if ((PINC | 0b11000000) == 0b11111110) {
			if (PIND == 0b10111111) {
				if (remap_counter >= 100) {
					button_mapping[6] = _BV(4);
					button_mapping[5] = _BV(5);
					button_mapping[7] = _BV(6);
					button_mapping[4] = _BV(7);
					button_mapping[1] = _BV(1);
					button_mapping[0] = _BV(3);
					button_mapping[3] = _BV(0);
					button_mapping[2] = _BV(2);
					remap_counter = 0;
				} else {
					++remap_counter;
				}
			} else if (PIND == 0b11011111) {
				if (remap_counter >= 100) {
					for (ix = 0, mask = 1; mask; ++ix, mask <<= 1) {
						button_mapping[ix] = mask;
					}
					remap_counter = 0;
				} else {
					++remap_counter;
				}
			} else {
				remap_counter = 0;
			}
		} else {
			remap_counter = 0;
		}
	}
}